from sidecar.status_maintainer import StatusMaintainer
from sidecar.const import Const
from sidecar.sandbox_error import SandboxError, ErrorCodes
from logging import Logger
from typing import List
import json
from sidecar.utils import Utils


class ErrorsController:
    def __init__(self,  status_maintainer: StatusMaintainer):
        self.status_maintainer = status_maintainer

    def add_sandbox_error(self, error: SandboxError):
        errors = self.status_maintainer.get_sandbox_errors()
        errors.append(error)
        self.status_maintainer.update_sandbox_errors(errors)

    def handle_artifacts_not_found(self, artifacts_not_found):
        if len(artifacts_not_found) > 0:
            error_msg = "Timeout reached while waiting for artifacts.\n" \
                        "Missing artifacts in your artifact storage provider:\n"
            index = 1
            for artifacts_name in artifacts_not_found:
                error_msg += "{INDEX}. {ARTIFACT_NAME}\n".format(INDEX=index, ARTIFACT_NAME=artifacts_name)
                index += 1

            error = SandboxError(Utils.get_utc_now_in_isoformat(), ErrorCodes.ARTIFACTS_DOWNLOAD_FAILED, error_msg)

            self.add_sandbox_error(error)




import threading
from datetime import datetime
from typing import List

from retrying import retry

from sidecar.azure_clp.data_store_service import DataStoreService
from sidecar.azure_clp.retrying_helpers import retry_if_connection_error, retry_if_result_none_or_empty
from sidecar.const import Const
from sidecar.status_maintainer import StatusMaintainer
from logging import Logger
from sidecar.sandbox_error import SandboxError




class AzureStatusMaintainer(StatusMaintainer):
    ROOT_SPEC_KEY = "spec"
    EXPECTED_APPS_KEY = "expected_apps"
    ROOT_APPS_KEY = "apps"
    INSTANCES_KEY = "instances"
    ROOT_SERVICES_KEY = "services"

    def __init__(self, data_store_service: DataStoreService, sandbox_id: str, logger: Logger):
        super().__init__(logger)
        self._data_store_service = data_store_service
        self._sandbox_id = sandbox_id
        self._lock = threading.RLock()

        self._cached_realtime_apps = None
        self._cached_realtime_services = None
        self._cached_spec = None
        self._cached_sandbox_errors = []

        self._cache_sandbox_data()

    def _cache_sandbox_data(self):
        sandbox_document = self._get_sandbox_document_with_retries()
        if not sandbox_document:
            raise Exception("Failed to get sandbox data from cosmos_db after 5 retries")

        self._cached_realtime_apps = sandbox_document[self.ROOT_APPS_KEY]
        self._cached_realtime_services = sandbox_document[self.ROOT_SERVICES_KEY]
        self._cached_spec = sandbox_document[self.ROOT_SPEC_KEY]
        # the get and [] is for backward compatibility (so that if there is a sandbox that don't have the
        # item SANDBOX_ERRORS in the DB we will return [] and not get Noun)
        self._cached_sandbox_errors = sandbox_document.get(Const.SANDBOX_ERRORS, [])

    @retry(stop_max_attempt_number=5, wait_fixed=1000, retry_on_result=retry_if_result_none_or_empty,
           retry_on_exception=retry_if_connection_error)
    def _get_sandbox_document_with_retries(self):
        return self._data_store_service.find_data_by_id(data_id=self._sandbox_id)

    def _construct_instances_key_path(self, instance_logical_id):
        return '{apps_key}.{instance_logical_id}.{instances_key}'.format(apps_key=self.ROOT_APPS_KEY,
                                                                         instance_logical_id=instance_logical_id,
                                                                         instances_key=self.INSTANCES_KEY)

    def update_app_instance_status(self, instance_logical_id, instance_id, app_name, status):
        with self._lock:
            instances_under_logical_id = self._cached_realtime_apps[instance_logical_id][self.INSTANCES_KEY]
            apps_under_instance_id = instances_under_logical_id.setdefault(instance_id, {"apps": {}})["apps"]
            app_details = apps_under_instance_id.setdefault(app_name, {})
            app_details[Const.APP_STATUS_TAG] = status
            self._cached_realtime_apps[instance_logical_id][self.INSTANCES_KEY] = instances_under_logical_id

            self._data_store_service.update_data(data_id=self._sandbox_id,
                                                 data=self._cached_realtime_apps,
                                                 column_name=self.ROOT_APPS_KEY)

    def get_app_names_on_instance(self, instance_logical_id: str) -> List[str]:
        return self._cached_spec[self.EXPECTED_APPS_KEY][instance_logical_id]["apps"]

    def update_sandbox_end_status(self, sandbox_deployment_end_status: str):
        # assume that DB-level lock is good enough here
        self._data_store_service.update_data(data_id=self._sandbox_id,
                                             data=sandbox_deployment_end_status,
                                             column_name=Const.SANDBOX_DEPLOYMENT_END_STATUS)

        self._data_store_service.update_data(data_id=self._sandbox_id,
                                             data=sandbox_deployment_end_status,
                                             column_name=Const.SANDBOX_DEPLOYMENT_END_STATUS_v2)

    def update_sandbox_start_status(self, sandbox_start_time: datetime):
        # assume that DB-level lock is good enough here

        # todo gil:add update data bulk
        self._data_store_service.update_data(data_id=self._sandbox_id,
                                             data=str(sandbox_start_time),
                                             column_name=Const.SANDBOX_START_TIME)

        self._data_store_service.update_data(data_id=self._sandbox_id,
                                             data=str(sandbox_start_time),
                                             column_name=Const.SANDBOX_START_TIME_v2)

    def update_service_status(self, name: str, status: str):
        self._cached_realtime_services[name]["status"] = status
        self._data_store_service.update_data(data_id=self._sandbox_id,
                                             data=self._cached_realtime_services,
                                             column_name=self.ROOT_SERVICES_KEY)

    def update_lazy_load_artifacts_waiting(self, value: bool):
        self._data_store_service.update_data(data_id=self._sandbox_id,
                                             data=bool(value),
                                             column_name=Const.LAZY_LOAD_ARTIFACTS_WAITING)

    def update_sandbox_errors(self, errors: List[SandboxError]):
        with self._lock:
            self._cached_sandbox_errors = [{"message": error.message, "code": error.code, "time": error.time} for error in errors]

            self._data_store_service.update_data(data_id=self._sandbox_id,
                                                 data=self._cached_sandbox_errors,
                                                 column_name=Const.SANDBOX_ERRORS)

    def get_sandbox_errors(self) -> List[SandboxError]:
        errors = []
        for error in self._cached_sandbox_errors:
            errors.append(SandboxError(message=str(error["message"]),
                                       code=str(error["code"]),
                                       time=str(error["time"])))
        return errors


    def update_service_outputs(self, name: str, outputs: {}):
        self._cached_realtime_services[name]["outputs"] = outputs
        self._data_store_service.update_data(data_id=self._sandbox_id,
                                             data=self._cached_realtime_services,
                                             column_name=self.ROOT_SERVICES_KEY)

    def get_service_deployment_outputs(self, service_name: str) -> {}:
        service = self._cached_realtime_services.get(service_name, None)
        if service:
            return service.get("outputs", {})
        raise Exception(f"could not find service with name '{service_name}'")

    def update_app_instance_deployment_outputs(self, instance_logical_id, instance_id, app_name, outputs: {}):
        with self._lock:
            instances_under_logical_id = self._cached_realtime_apps[instance_logical_id][self.INSTANCES_KEY]
            apps_under_instance_id = instances_under_logical_id.setdefault(instance_id, {"apps": {}})["apps"]

            app_details = apps_under_instance_id.setdefault(app_name, {})
            app_details['outputs'] = outputs
            self._cached_realtime_apps[instance_logical_id][self.INSTANCES_KEY] = instances_under_logical_id

            self._data_store_service.update_data(data_id=self._sandbox_id,
                                                 data=self._cached_realtime_apps,
                                                 column_name=self.ROOT_APPS_KEY)

    def get_app_deployment_outputs(self, app_name: str) -> {}:
        app_details = self._get_first_application(app_name)
        if app_details:
            return app_details.get('outputs', {})
        raise Exception(f"could not find application with name '{app_name}'")

    def _get_first_application(self, app_name: str):
        for logical_id, logical_details in self._cached_realtime_apps.items():
            for instance_id, instance_json in logical_details[self.INSTANCES_KEY].items():
                for name, app_details in instance_json['apps'].items():
                    if name == app_name:
                        return app_details
        return None



class SandboxError:
    def __init__(self, time: str, code: str, message: str):
        self.message = message
        self.code = code
        self.time = time


class ErrorCodes:
    ARTIFACTS_DOWNLOAD_FAILED = "ARTIFACTS_DOWNLOAD_FAILED"



from typing import List, Tuple

from sidecar.apps_configuration_end_tracker import AppsConfigurationEndTracker, AppConfigurationEndStatus
from sidecar.const import AppNetworkStatus
from sidecar.health_check.app_health_check_state import AppHealthCheckState
from sidecar.model.objects import SidecarApplication, ISidecarService
from sidecar.sandbox_terminating_query import ISandboxTerminatingQuery
from sidecar.services.service_status_state import ServiceStatusState


class ConfigurationStartStatus:
    START = "start"
    WAIT = "wait"
    CANCEL = "cancel"


class ConfigurationStartPolicy:
    def __init__(self,
                 service_status_state: ServiceStatusState,
                 app_health_check_state: AppHealthCheckState,
                 apps_config_end_tracker: AppsConfigurationEndTracker,
                 apps: List[SidecarApplication],
                 services: List[ISidecarService],
                 terminating_query: ISandboxTerminatingQuery = None):
        self._service_status_state = service_status_state
        self._services = services
        self._apps = apps
        self._app_health_check_state = app_health_check_state
        self._apps_config_end_tracker = apps_config_end_tracker

        self._app_dependencies = {app.name: list(app.dependencies) for app in apps}
        self._service_dependencies = {service.name: list([d for d in service.dependencies]) for service in services}

        self._dependencies = {**self._app_dependencies, **self._service_dependencies}

        self._terminating_query = terminating_query

    def get_configuration_start_status(self, name: str) -> str:
        if self._terminating_query and self._terminating_query.is_terminating():
            return ConfigurationStartStatus.CANCEL

        dependencies = self._dependencies[name]
        if not dependencies:
            return ConfigurationStartStatus.START

        app_dependencies, service_dependencies = self.split_dependencies_by_type(dependencies)

        app_dependency_statuses = self._apps_config_end_tracker.get_app_configuration_statuses(*app_dependencies)
        all_instances_completed_with_success = all(
            config_end_status.is_ended_with_status(AppConfigurationEndStatus.COMPLETED) for config_end_status in
            app_dependency_statuses.values())

        all_apps_completed_private_network_health_check = all(
            AppNetworkStatus.passed_internal_network_test(status)
            for status
            in self._app_health_check_state.get_apps_state(app_dependencies).values())

        all_service_dependencies_ended_with_success = \
            self._service_status_state.execution_done(names=service_dependencies)

        if all_instances_completed_with_success and \
                all_apps_completed_private_network_health_check and \
                all_service_dependencies_ended_with_success:
            return ConfigurationStartStatus.START

        return ConfigurationStartStatus.WAIT

    def split_dependencies_by_type(self, dependencies: List[str]) -> Tuple[List[str], List[str]]:
        apps = []
        services = []
        for d in dependencies:
            if d in self._app_dependencies:
                apps.append(d)
            if d in self._service_dependencies:
                services.append(d)
        return apps, services
